@@ -1,13 +1,15 @@
+from django.conf import settings
 from django.db import models
 from django.contrib.contenttypes.models import ContentType
 from django.contrib.contenttypes import generic
 
+from django.utils.timesince import timesince as _timesince
 from django.utils.translation import ugettext as _
 
-from activity.compat import User
 from activity.registry import activityregistry
 from activity.signals import action
-from activity.managers import ActionManager, FollowManager
+from activity.managers import ActionManager, FollowManager, StreamManager
+from activity.tasks import fanout_action
 
 
 class Action(models.Model):
@@ -16,7 +18,6 @@ class Action(models.Model):
 
     Naming convention from http://activitystrea.ms/specs/atom/1.0/
     """
-
     handler = models.CharField(max_length=255)
 
     actor_content_type = models.ForeignKey(ContentType, related_name='actor')
@@ -33,6 +34,7 @@ class Action(models.Model):
 
     created = models.DateTimeField(auto_now_add=True)
     public = models.BooleanField(default=True)
+    is_global = models.BooleanField(default=False)
 
     objects = ActionManager()
 
@@ -42,30 +44,61 @@ class Action(models.Model):
     def __unicode__(self):
         values = {
             'actor': self.actor,
-            'handler': self.handler,
+            'verb': self.verb,
+            'action_object': 'joo',
             'target': self.target,
             'since': self.timesince()
         }
-        return _('%(handler)s by %(actor)s, %(target)s [%(since)s ago]') % values
+        if self.target:
+            if self.action_object:
+                return _('%(actor)s %(verb)s %(action_object)s on %(target)s %(since)s ago') % values
+            else:
+                return _('%(actor)s %(verb)s %(target)s %(since)s ago') % values
+        if self.action_object:
+            return _('%(actor)s %(verb)s %(action_object)s %(since)s ago') % values
+        return _('%(actor)s %(verb)s %(since)s ago') % values
+
+    @property
+    def verb(self):
+        """
+        Get action's verb
+        """
+        handlers = activityregistry.get_handlers()
+        handler = handlers[self.handler]
+        return handler.verb
 
     def timesince(self, now=None):
         """
         Shortcut for ``django.utils.timesince.timesince`` function
         """
-        from django.utils.timesince import timesince as _timesince
         return _timesince(self.created, now)
 
+
+class Stream(models.Model):
+    """
+    User's activity stream item. This is used to pre-populate activity streams
+    and to avoid scans on Action table.
+    """
+    user = models.ForeignKey(settings.AUTH_USER_MODEL)
+    action = models.ForeignKey(Action)
+
+    objects = StreamManager()
+
+    class Meta:
+        unique_together = ('user', 'action')
+
+
 class Follow(models.Model):
     """
     Let user to follow activities of any user or object
     """
-    user = models.ForeignKey(User)
+    user = models.ForeignKey(settings.AUTH_USER_MODEL)
 
     # Object to Follow
     content_type = models.ForeignKey(ContentType)
     object_id = models.PositiveIntegerField()
     follow_object = generic.GenericForeignKey()
-    actor_only = models.BooleanField('Only follow actions where the object is the target', default=True)
+    actor_only = models.BooleanField('Only follow actions where the object is the actor', default=True)
 
     started = models.DateTimeField(auto_now_add=True)
 
@@ -83,11 +116,13 @@ def action_handler(sender, **kwargs):
     handlers = activityregistry.get_handlers()
     handler_name = kwargs.get('handler')
     actor = kwargs.get('actor')
+    is_global = kwargs.get('is_global', False)
     if handler_name in handlers:
         action = Action(
             handler=handler_name,
             actor_content_type=ContentType.objects.get_for_model(actor),
             actor_object_id=actor.pk,
+            is_global=is_global,
         )
         for opt in ('action_object', 'target'):
             obj = kwargs.get(opt, None)
@@ -97,5 +132,8 @@ def action_handler(sender, **kwargs):
                         ContentType.objects.get_for_model(obj))
         action.save()
 
+        # Fanout action (populate streams)
+        fanout_action.delay(action.pk)
+
 
 action.connect(action_handler, dispatch_uid='activity.models')
